#!/usr/bin/perl

use warnings;
use strict;

use Data::Dumper;

my $in_file= "MULE - Disassembled Memory.asm.txt";

## -- Read source --

my $fin;
open $fin, $in_file or die "Can't read $in_file: $!";
my @line= <$fin>;
close $fin;

chomp $_ for (@line);

## -- Scan for possible labels --

my %label= ();
my %label_used= ();
my %label_is_sub= ();
my $inside_sub= '';
my %line_sub= ();
for my $line (@line) {

    if ($line =~ / S U B R O U T I N E /) {
        if($inside_sub) {
            die "End of function $inside_sub not found";
        }
        $inside_sub= ':';
    }

    if ($line =~ /^[^;]*\b([a-zA-Z][A-Za-z0-9_]+):/) {
        my $label= $1;
        if ($label !~ /^for$/) {
            $line =~ s/^\s+//;
            $label{$label}= [ $line ];
            $label_used{$label}= [];
            if ($inside_sub eq ':') {
                $inside_sub= $label;
                $line_sub{$line}= $inside_sub;
                $label_is_sub{$inside_sub}= 1;
            }
        }
    }

    $line_sub{$line}= $inside_sub if $inside_sub && $inside_sub ne ':';

    if ($line =~ /; End of function $inside_sub/) {
        $inside_sub= '';
    }
}

my $mouse_out= "onmouseout=\"parent.ov()\"";

$mouse_out= ""; # uncomment to turn off mouseout

# print Dumper(\%label);

## -- Collect label usage, replace with link --

sub rpl_link {
    my $label= shift;
    my $in_comment= shift;
    my $line= shift;
    return $label unless defined $label{$label};
    my $n= scalar @{ $label{$label} } - 1;
    $line =~ s/^\s+//;
    push @{ $label{$label} }, $line;
    $label_used{$label}++ unless $in_comment;
    my $pos= length($label) >= 4 ? 2 : 1;
    return "<a id=\"$label-$n-xref\" target=\"xref\" href=\"xref.html#$label\" onmouseover=\"parent.ov('$label')\" $mouse_out>" . substr($label, 0, $pos) . "</a>"
        . "<a id=\"$label-$n\" name=\"$label-$n\" href=\"#$label\" onmouseover=\"parent.ov('$label')\" $mouse_out>" . substr($label, $pos) . "</a>";
}

for my $line (@line) {
    $line=~ /^(.*?)(;.*)?$/;
    my $code= $1;
    my $comment= $2 || '';
    $code =~    s/\b([a-zA-Z][A-Za-z0-9_]+)\b(?!:)/rpl_link($1, 0, $line)/ge;
    $comment =~ s/\b([a-zA-Z][A-Za-z0-9_]+)\b/rpl_link($1, 1, $line)/ge;
    $line= "$code$comment";
}

## -- Replace label definition with link --

sub rpl_label {
    my $label= shift;
    return $label unless defined $label_used{$label} && $label_used{$label};
    return "<strong><a id=\"$label\" name=\"$label\" target=\"xref\" href=\"xref.html#$label\" onmouseover=\"parent.ov('$label')\" $mouse_out>$label</a></strong>";
}

for my $line (@line) {
    $line =~ s/^([^;]*)\b([a-zA-Z][A-Za-z0-9_]+):/$1 . rpl_label($2) . ':'/e;
}

## -- Output result --

my @contents;

mkdir "out";

sub write_file {
    my $file_name= shift;
    my $contents= shift;
    my $no_pre= shift || 0;

    $contents= join("\n", @{ $contents }) if ref($contents);

    my $fout;
    open $fout, ">out/$file_name.html" or die "Can't write $file_name.html: $!";
    print $fout "<pre>\n" unless $no_pre;
    print $fout $contents;
    print $fout "</pre>\n" unless $no_pre;
    close $fout;
}

## -- Write index.html --

write_file("index", q(
    <script>
        var ovEl;
        var _ov= function(className) {
            ovEl.className= className;
            for (var i= 0;; i++) {
                var ovEl2= src.document.getElementById(ovEl.id + '-' + i);
                if (ovEl2 == null) break;
                ovEl2.className= className + 'ref';
                ovEl2= src.document.getElementById(ovEl.id + '-' + i + '-xref');
                if (ovEl2 != null) ovEl2.className= className;
            }
        };
        var ov= function(labelId) {
            if (ovEl) _ov('');
            ovEl= src.document.getElementById(labelId);
            if (ovEl) _ov('ov');
        };
    </script>
    <title>
        ) . $in_file . q(
    </title>
    <frameset cols="800,*">
        <frameset rows="30,*,*">
            <frame name="labels_nav" src="labels_nav.html"> 
            <frame name="labels" src="labels_id_by_name.html"> 
            <frame name="xref" src="xref.html"> 
        </frameset>
        <frame name="src" src="src.html"> 
    </frameset>
), 1);

## -- Write labels_by_name.html and labels_by_adr.html --

our $reg= qr/^(word_|unk_|sub_|OS_|off_|locret_|loc_|byte_)/;

sub label_sort {
    my $au= $a =~ $reg; # /^(word_|unk_|sub_|OS_|off_|locret_|loc_|byte_|v\d)/;
    my $bu= $b =~ $reg; # /^(word_|unk_|sub_|OS_|off_|locret_|loc_|byte_|v\d)/;
    return 1  if $au && !$bu;
    return -1 if !$au && $bu;
    return lc($a) cmp lc($b);
}

@contents= ();
my $last_label= '';
my $id_by_name_written= 0;
for my $label (sort label_sort keys %label) {
    next unless $label_used{$label};
    if ($label =~ $reg && $last_label ne $1) {

        write_file("labels_id_by_name", \@contents) unless $id_by_name_written;
        $id_by_name_written= 1;

        push @contents, "<hr>";
        $last_label= $1;
    }
    my $prefix= sprintf("%4s ", (defined($label_is_sub{$label}) ? "SUB" : "")) . substr($label{$label}[0], 0, 4);
    push @contents, "$prefix <a target=\"src\" href=\"src.html#$label\" onmouseover=\"parent.ov('$label')\" $mouse_out>$label</a>";
}
write_file("labels_by_name", \@contents);

## -- Write labels_id_by_name.html --

sub label2_sort {
    return $label{$a}[0] cmp $label{$b}[0];
}

@contents= ();
for my $label (sort label2_sort keys %label) {
    next unless $label_used{$label};
    next if $label =~ $reg;
    my $prefix= sprintf("%4s ", (defined($label_is_sub{$label}) ? "SUB" : "")) . substr($label{$label}[0], 0, 4);
    push @contents, "$prefix <a target=\"src\" href=\"src.html#$label\" onmouseover=\"parent.ov('$label')\" $mouse_out>$label</a>";
}
write_file("labels_id_by_adr", \@contents);

## -- Write labels_nav.html --

@contents= ();
push @contents, "<a target=\"labels\" href=\"labels_id_by_name.html\">IdByName</a>"
    . " <a target=\"labels\" href=\"labels_id_by_adr.html\">IdByAdr</a>"
    . " <a target=\"labels\" href=\"labels_by_name.html\">AllByName</a>"
;
write_file("labels_nav", \@contents);

## -- Write xref.html --

# print Dumper(\%line_sub); die;

@contents= ();
for my $label (sort label_sort keys %label) {
    next unless $label_used{$label};
    # push @contents, "<strong>XREF <a name=\"$label\" target=\"src\" href=\"src.html#$label\" onmouseover=\"parent.ov('$label')\" $mouse_out>$label:</a></strong>";
    push @contents, "<a name=\"$label\"></a>";
    my $n= -1;
    my $first= 1;
    for my $line (@{ $label{$label} }) {
        my $descr= "";
        my $prefix= sprintf("%30s", defined($line_sub{$line}) ? $line_sub{$line} . ": " : "");
        if (!$first) { $descr= substr($line, 4); $line= substr($line, 0, 4); }
        $first= 0;
        push @contents, "$prefix<a target=\"src\" href=\"src.html#$label" . ($n >= 0 ? "-$n" : "") . "\" onmouseover=\"parent.ov('$label')\" $mouse_out>$line</a>$descr";
        $n++;
    }
    push @contents, "<br>";
}
write_file("xref", \@contents);

## -- Write src.html --

write_file("src", [ q(
        <style>
            .ov { background-color: #FF0; }
            .ovref { background-color: #0F0; }
        </style>
    ), @line ]);

## -- EOF --
